﻿if ( !Array.prototype.forEach ) {
	Array.prototype.forEach = function( callback, thisArg ) {
		var T, k;
		if ( this == null ) {
			throw new TypeError( " this is null or not defined" );
		}
		// 1. Let O be the result of calling ToObject passing the |this| value as the argument.
		var O = Object(this);
		// 2. Let lenValue be the result of calling the Get internal method of O with the argument "length".
		// 3. Let len be ToUint32(lenValue).
		var len = O.length >>> 0; // Hack to convert O.length to a UInt32
		// 4. If IsCallable(callback) is false, throw a TypeError exception.
		// See: http://es5.github.com/#x9.11
		if ( {}.toString.call(callback) != "[object Function]" ) {
			throw new TypeError( callback + " is not a function" );
		}
		// 5. If thisArg was supplied, let T be thisArg; else let T be undefined.
		if ( thisArg ) {
			T = thisArg;
		}
		// 6. Let k be 0
		k = 0;

		// 7. Repeat, while k < len
		while( k < len ) {
			var kValue;
			// a. Let Pk be ToString(k).
			//   This is implicit for LHS operands of the in operator
			// b. Let kPresent be the result of calling the HasProperty internal method of O with argument Pk.
			//   This step can be combined with c
			// c. If kPresent is true, then
			if ( k in O ) {
				// i. Let kValue be the result of calling the Get internal method of O with argument Pk.
				kValue = O[ k ];
				// ii. Call the Call internal method of callback with T as the this value and
				// argument list containing kValue, k, and O.
				callback.call( T, kValue, k, O );
			}
			// d. Increase k by 1.
			k++;
		}
		// 8. return undefined
	};
}

if (Object.assign === undefined) {
	Object.prototype.assign = function(target, source) {
//		for (var source of sources) {
			var keys = Object.getOwnPropertyNames(source);
			for (var p in keys) {
				var key = keys[p];
				var desc = Object.getOwnPropertyDescriptor(source, key);
				if (desc && desc.enumerable) {
					target[key] = source[key];
				}
			}
//		}
	};
}

function UI_Rect(x, y, w, h) {
	var rect = Object.create(UI_Rect);
	rect.set(x, y, w, h);
	return rect;
}

UI_Rect.isinside = function(x, y) {
	x = +x || 0;
	y = +y || 0;
	if (this.x <= x && x < this.x + this.w && this.y <= y && y < this.y + this.h) {
		return true;
	}
	else {
		return false;
	}
};

UI_Rect.isempty = function() {
	return this.w <= 0 || this.h <= 0;
}

UI_Rect.set = function(x, y, w, h) {
	this.x = +x || 0;
	this.y = +y || 0;
	this.w = +w || 0;
	this.h = +h || 0;
	return this;
}

UI_Rect.copy = function(r) {
	if (r) {
		this.x = +r.x || 0;
		this.y = +r.y || 0;
		this.w = +r.w || 0;
		this.h = +r.h || 0;
	}
	return this;
}

UI_Rect.or = function(r) {
	if (r) {
		r.x = +r.x || 0;
		r.y = +r.y || 0;
		r.w = +r.w || 0;
		r.h = +r.h || 0;
		if (!this.isempty() && r.w > 0 && r.h > 0) {
			let maxx = r.x + r.w > this.x + this.w ? r.x + r.w : this.x + this.w;
			let maxy = r.y + r.h > this.y + this.h ? r.y + r.h : this.y + this.h;
			this.x = r.x < this.x ? r.x : this.x;
			this.y = r.y < this.y ? r.y : this.y;
			this.w = maxx - this.x;
			this.h = maxy - this.y;
		}
	}
	return this;
}

UI_Rect.isequal = function(r) {
	if (r && r.x && r.y && r.w && r.h) {
		if (this.isempty() && (r.w <= 0 || r.h <= 0)) {
			return true;
		}
		else {
			return this.x == r.x && this.y == r.y && this.w == r.w && this.h == r.h;
		}
	}
	else {
		return false;
	}
}

UI_Rect.ctxrect = function(ctx) {
	ctx.rect(this.x + ctx.lineWidth / 2, this.y + ctx.lineWidth / 2, this.w - ctx.lineWidth, this.h - ctx.lineWidth);
}

function UI_Style(param) {
	var style = Object.create(UI_Style);
	style.font	=					(param && param.font)					||	"Century, sans serif";
	style.fontsize =			(param && param.fontsize)			|| 12;
	style.textStyle =			(param && param.textStyle)		|| "black";
	style.strokeStyle =		(param && param.strokeStyle)	|| "black";
	style.lineWidth =			(param && param.lineWidth)		|| 0;
	style.fillStyle =			(param && param.fillStyle)		|| "transparent";
	style.align =					(param && param.align)				|| "left";
	style.valign =				(param && param.valign)				|| "middle";
	style.px =						(param && param.px)						|| 0;
	style.py =						(param && param.py)						|| 0;
	style.display =				(param && param.display !== undefined) ? param.display : true;
	return style;
}

UI_canvas = document.createElement("canvas");
UI_ctx = UI_canvas.getContext('2d');

function UI_Drawdata(type, param) {
	var drawdata =						Object.create(UI_Drawdata);
	drawdata.type =						type;
	drawdata.id =							(param && param.id !== undefined) ? param.id : drawdata.type;
	drawdata.value =					(param && param.value !== undefined) ? param.value : null;
	drawdata.style =					UI_Style(param && param.style);
	drawdata.moverstyle =			(param && param.moverstyle) ? UI_Style(param.moverstyle) : drawdata.style;
	drawdata.offstyle =				(param && param.offstyle) ? UI_Style(param.offstyle) : drawdata.style;
	drawdata.moveroffstyle =	(param && param.moveroffstyle) ? UI_Style(param.moveroffstyle) : drawdata.moverstyle;
	drawdata.disabledstyle =	(param && param.disabledstyle) ? UI_Style(param.disabledstyle) : drawdata.style;
	drawdata.x =							(param && param.x) || 0;
	drawdata.y = 							(param && param.y) || 0;
	drawdata.width =					(param && param.width) || 50;
	drawdata.height	=					(param && param.height) || 50;
	drawdata.position =				(param && param.position) || "static";
	drawdata.transform =			(param && param.transform) || null;
	drawdata.param =					(param && param.param) || null;
	drawdata.display =				(param && param.display !== undefined)? param.display : true;
	drawdata.rect =						UI_Rect(drawdata.x, drawdata.y, drawdata.width, drawdata.height);
	drawdata.shortcutkey =		(param && param.shortcutkey) || "";
	drawdata.tooltip =				(param && param.tooltip) || null;
	drawdata.tooltiptime =		(param && param.tooltiptime) || 60;
	drawdata.mouseover = false;
	drawdata.on	= true;
	drawdata.mouseovercounter = 0;
	return drawdata;
}

UI_Drawdata.prototype.gettype = function() {
	return this.type;
},
UI_Drawdata.prototype.update = function() {
	this.rect =		UI_Rect(this.rect.x, this.rect.y, this.width, this.height);
}
UI_Drawdata.prototype.getstyle = function() {
	if (this.disabled) {
		return this.disabledstyle;
	}
	return (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
}
UI_Drawdata.prototype.drawcommon = function(ctx) {
	var style = this.getstyle();
	if (!style.display) {
		return;
	}
	this.drawbackground(ctx);
	this.draw(ctx);
}
UI_Drawdata.prototype.drawbackground = 	function(ctx) {
	var style = this.getstyle();
	ctx.beginPath();
	ctx.rect(this.rect.x, this.rect.y, this.rect.w, this.rect.h);
	ctx.fillStyle = style.fillStyle;
	ctx.fill();
	if (style.lineWidth > 0) {
		ctx.beginPath();
		ctx.lineWidth		= style.lineWidth;
		ctx.strokeStyle	= style.strokeStyle;
		ctx.rect(this.rect.x + style.lineWidth / 2, this.rect.y + style.lineWidth / 2, this.rect.w - style.lineWidth, this.rect.h - style.lineWidth);
		ctx.strokeStyle	= style.strokeStyle;
		ctx.stroke();
	}
}

function UI_Drawdata_Text(param) {
	var ddata = Object.create(UI_Drawdata_Text.prototype);
	Object.assign(ddata, UI_Drawdata("text", param));
	ddata.text				= (param && param.text) || "";
	ddata.autosize		= (param && param.autosize) || false;
	ddata.bindvar			= (param && param.bindvar) || null;
	return ddata;
}

UI_Drawdata_Text.prototype.update =	function() {
	let width = this.width;
	if (this.autosize) {
		var style = this.getstyle();
		let ctx = UI_ctx;
		ctx.fillStyle			= style.textStyle;
		ctx.textBaseline	= "bottom";
		ctx.font					= style.fontsize + "px " + style.font;
		var text = this.text = (this.bindvar) ? eval(this.bindvar) : this.text;
		width = ctx.measureText(text).width;
	}
	this.rect =		UI_Rect(this.rect.x, this.rect.y, width, this.height);
}

UI_Drawdata_Text.prototype.draw =	function(ctx) {
	var style = this.getstyle();
	ctx.fillStyle			= style.textStyle;
	ctx.textBaseline	= "bottom";
	ctx.font					= style.fontsize + "px " + style.font;
	var text = this.text = (this.bindvar) ? eval(this.bindvar) : this.text;
	var twidth = ctx.measureText(text).width;
	var x = this.rect.x + style.lineWidth + style.px;
	var y = this.rect.y + style.lineWidth + style.fontsize + style.py;
	if (style.align === "center") {
		x = this.rect.x + (this.rect.w - twidth) / 2;
	}
	else if (style.align === "right") {
		x = this.rect.x + this.rect.w - twidth - style.lineWidth - style.px;
	}
	if (style.valign === "middle") {
		y = this.rect.y + (style.fontsize + this.rect.h) / 2 + 1;
	}
	else if (style.valign === "bottom") {
		y = this.rect.y + this.rect.h - style.lineWidth - style.py;
	}
	x |= 0;
	y |= 0;
	ctx.fillText(text, x, y);
}

Object.setPrototypeOf(UI_Drawdata_Text.prototype, UI_Drawdata.prototype);

function UI_Drawdata_Image(param) {
	var ddata = Object.create(UI_Drawdata_Image.prototype);
	Object.assign(ddata, UI_Drawdata("image", param));
	if (param && param.image) {
		ddata.image = param.image;
	}
	else {
		ddata.image	=				new Image();
		if (param && param.src) {
			ddata.image.src =		(param && param.src) || null;
			ddata.image.onload = function(param) {
				this.imagewidth =		this.imagewidth || this.image.width;
				this.imageheight =	this.imageheight || this.image.height;
			}.bind(ddata);
		}
	}
	ddata.imagewidth =	(param && param.imagewidth) || ddata.image.width;
	ddata.imageheight =	(param && param.imageheight) || ddata.image.height;

	ddata.imagesrcx		=	(param && param.imagesrcx) || 0;
	ddata.imagesrcy		=	(param && param.imagesrcy) || 0;
	ddata.imagesrcw		=	(param && param.imagesrcw) || 0;
	ddata.imagesrch		=	(param && param.imagesrch) || 0;
	return ddata;
}

UI_Drawdata_Image.prototype.draw = function(ctx) {
	var style = this.getstyle();
	var x = this.rect.x + style.lineWidth + style.px;
	var y = this.rect.y + style.lineWidth + style.py;
	let w = this.imagewidth =		this.imagewidth || this.image.width;
	let h = this.imageheight =	this.imageheight || this.image.height;
	let sw = w;
	let sh = h;
	if (this.imagesrcw) {
		sw = this.imagesrcw;
		sh = this.imagesrch;
	}
	if (style.align === "center") {
		x = this.rect.x + (this.rect.w - w) / 2;
	}
	else if (style.align === "right") {
		x = this.rect.x + this.rect.w -  w - style.lineWidth - style.px;
	}
	if (style.valign === "middle") {
		y = this.rect.y + (this.rect.h - h) / 2;
	}
	else if (style.valign === "bottom") {
		y = this.rect.y + this.rect.h - h - style.lineWidth - style.py;
	}
	if (this.transform) {
		ctx.save();
		if (this.transform === "reversex") {
			ctx.scale(-1, 1);
			ctx.translate(-x * 2 - w, 0);
		}
	}
	x |= 0;
	y |= 0;
	if (this.imagesrcw) {
		ctx.drawImage(this.image, this.imagesrcx, this.imagesrcy, sw, sh, x, y, w, h);
	}
	else {
		ctx.drawImage(this.image, x, y, w, h);
	}
	if (this.transform) {
		ctx.restore();
	}
}

Object.setPrototypeOf(UI_Drawdata_Image.prototype, UI_Drawdata.prototype);

function UI_Drawdata_PlusIntegerImage(param) {
	var ddata = Object.create(UI_Drawdata_PlusIntegerImage.prototype);
	Object.assign(ddata, UI_Drawdata("plusintegerimage", param));
	if (param && param.image) {
		ddata.image = param.image;
	}
	else {
		ddata.image	=			[];
		if (param && param.src) {
			ddata.image = new Image();
			ddata.image[i].src =	param.src;
		}
		ddata.image.onload = function() {
			this.imagewidth =		this.imagewidth || this.image[0].width;
			this.imageheight =	this.imageheight || this.image[0].height;
		}.bind(ddata);
	}
	ddata.text = (param && param.text) || 0;
	ddata.imagewidth =	(param && param.imagewidth) || (ddata.image && ddata.image[0] && ddata.image[0].width) || 0;
	ddata.imageheight =	(param && param.imageheight) || (ddata.image && ddata.image[0] && ddata.image[0].height) || 0;
	ddata.imagesrcx		=	(param && param.imagesrcx) || 0;
	ddata.imagesrcy		=	(param && param.imagesrcy) || 0;
	ddata.imagesrcw		=	(param && param.imagesrcw) || 0;
	ddata.imagesrch		=	(param && param.imagesrch) || 0;
	return ddata;
}

UI_Drawdata_PlusIntegerImage.prototype.draw = function(ctx) {
	var style = this.getstyle();
	var x = this.rect.x + style.lineWidth + style.px;
	var y = this.rect.y + style.lineWidth + style.py;

	var num = Math.floor(+this.text);
	if (!isFinite(num) || num < 0) {
		num = 0;
	}
	this.text = num + '';
	let imagewidth = this.imagesrcw * this.text.length;

	if (style.align === "center") {
		x = this.rect.x + (this.rect.w - imagewidth) / 2;
	}
	else if (style.align === "right") {
		x = this.rect.x + this.rect.w -  imagewidth - style.lineWidth - style.px;
	}
	if (style.valign === "middle") {
		y = this.rect.y + (this.rect.h - this.imagesrch) / 2;
	}
	else if (style.valign === "bottom") {
		y = this.rect.y + this.rect.h - this.imagesrch - style.lineWidth - style.py;
	}
	if (this.transform) {
		ctx.save();
		if (this.transform === "reversex") {
			ctx.scale(-1, 1);
			ctx.translate(-x * 2 - imagewidth, 0);
		}
	}
	x |= 0;
	y |= 0;
	for (let i = 0 ; i < this.text.length ; i++) {
		ctx.drawImage(this.image, this.imagesrcx + (this.text.substr(i, 1) * this.imagesrcw), this.imagesrcy, this.imagesrcw, this.imagesrch, x + this.imagesrcw * i, y, this.imagesrcw, this.imagesrch);
	}
	if (this.transform) {
		ctx.restore();
	}
}

Object.setPrototypeOf(UI_Drawdata_PlusIntegerImage.prototype, UI_Drawdata.prototype);

UI_CreateDrawdata = function(common, data) {
	let retvalue = [];
	let maxnum = 0;
	for (let name in data) {
		if (Object.prototype.toString.call(data[name]) === "[object Array]") {
			if (data[name].length > maxnum) {
				maxnum = data[name].length;
			}
		}
	}
	for (let i = 0 ; i < maxnum ; i++) {
		for (let name in data) {
			if (Object.prototype.toString.call(data[name]) === "[object Array]") {
				if (data[name][i] !== undefined) {
					common[name] = data[name][i];
				}
			}
		}
		let type = common.type;
		if (type === undefined) {
			type = "text";
		}
		if (type === "text") {
			retvalue.push(UI_Drawdata_Text(common));
		}
		if (type === "image") {
			retvalue.push(UI_Drawdata_Image(common));
		}
		if (type === "plusintegerimage") {
			retvalue.push(PlusIntegerImage(common));
		}
	}
	return retvalue;
}


UI2.eventlist = [
	"mousedown",
	"mouseup",
	"mousemove",
	"mouseout",
	"keydown",
	"keyup",
	"wheel",
];

function UI2(type, param) {
	var ui = Object.create(UI2.prototype);
	ui.type =						type;
	ui.id =							(param && param.id !== undefined) ? param.id : ui.type;
	ui.value =					(param && param.value !== undefined) ? param.value : null;
	ui.x =							(param && param.x) || 0;
	ui.y = 							(param && param.y) || 0;
	ui.px =							(param && param.px) || 0;
	ui.py = 						(param && param.py) || 0;
	ui.param = 					(param && param.param !== undefined) ? param.param : null;
	ui.position = 			(param && param.position) || "static";
	ui.width	=					(param && param.width) || 0;
	ui.height	=					(param && param.height) || 0;
	ui.autosize =				(param && param.autosize !== undefined) ? param.autosize : true;		// true: 自動、"w": 幅のみ自動、"h": 高さのみ自動
	ui.style =					(param && param.style) ? UI_Style(param.style) : null;
	ui.moverstyle =			(param && param.moverstyle) ? UI_Style(param.moverstyle) : ui.style;
	ui.offstyle =				(param && param.offstyle) ? UI_Style(param.offstyle) : ui.style;
	ui.moveroffstyle =	(param && param.moveroffstyle) ? UI_Style(param.moveroffstyle) : ui.moverstyle;
	ui.disabledstyle =	(param && param.disabledstyle) ? UI_Style(param.disabledstyle) : ui.style;
	ui.arg =						(param && param.arg) || null;
	ui.canvas =					(param && param.canvas) || null;
	ui.ctx =						(param && param.ctx) || null;
	ui.data	=						(param && param.data) || [];
	ui.display =				(param && param.display !== undefined) ? param.display : true;
	ui.disabled =				(param && param.disabled !== undefined) ? param.disabled : false;
	ui.enablewheel =		(param && param.enablewheel) || false;
	ui.enablescrollwheel =	(param && param.enablescrollwheel !== undefined) ? param.enablescrollwheel : true;
	ui.enableshortcut = (!param || !param.enableshortcut) ? true : false;
	ui.shortcutkey =		(param && param.shortcutkey) || "";
	ui.enabletooltip = 	(!param || !param.enabletooltip) ? true : false;
	ui.tooltip =				(param && param.tooltip) || null; 
	ui.tooltiptime =		(param && param.tooltiptime) || 60;
	ui.scrollbar	=			(param && param.scrollbar) || false;	// false: なし、 true:	両方あり、"v": 垂直のみあり、 "h":	水平のみあり、"auto": 自動設定
	ui.name =						(param && param.name) || "";
	ui.rect =						UI_Rect();
	ui.rectall = 				UI_Rect();
	ui.scrollbar_data = {
		v:	{ top:	0,	width: (param && param.scrollbar_vwidth) || 15, rect: UI_Rect(), tabrect: UI_Rect(), wheely: (param && param.scrollbar_wheely) || 10 },
		h:	{ left:	0,	height: (param && param.scrollbar_hheight) || 15, rect: UI_Rect(), tabrect: UI_Rect(), wheelx: (param && param.scrollbar_wheelx) || 10 },
	};
	if (param && param.preprocess) {
		ui.preprocess = param.preprocess;
	}
	if (param && param.bindvar) {
		ui.bindvar = param.bindvar;
	}
	if (param && param.onchange) {
		ui.onchange = param.onchange;
	}
	UI2.eventlist.forEach(function(data, i, a) {
		if (param && param["on" + data]) {
			ui["on" + data] = param["on" + data];
		}
	});
	if (param && param.onoperatedbyuser) {
		ui.onoperatedbyuser = param.onoperatedbyuser;
	}

	if (ui.data.forEach === undefined) {
		ui.data = [ ui.data ];
	}
	let newdata = [];
	ui.data.forEach(function(data) {
		if (Object.prototype.toString.call(data) === "[object Array]") {
			newdata = newdata.concat(data);
		}
		else {
			newdata.push(data);
		}
	});
	ui.data = newdata;
	ui.mouseover =	false;
	ui.isfirst = true;
	ui.mouseovercounter = 0;
	ui.mousedownflag = false;
	return ui;
}

UI2.prototype.getstyle = function() {
	if (this.disabled) {
		return this.disabledstyle;
	}
	return (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
}

UI2.prototype.update = function() {
//	this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	this.data.forEach(function(data, i, a) {
		data.rect.x =		this.x + data.x;
		data.rect.y =		this.y + data.y;
		data.arg =			this.arg;
		data.base =			this.base;
		data.update();
	}, this);
}
UI2.prototype.setrect = function(allw, allh) {
	this.rect.w = this.width;
	this.rect.h = this.height;
	this.rectall.copy(this.rect);
	this.rectall.w = allw;
	this.rectall.h = allh;
	if (this.autosize === true || this.autosize === "width") {
		this.rect.w = allw;
	}
	if (this.autosize === true || this.autosize === "height") {
		this.rect.h = allh;
	}
	// 縦、横スクロールバーの有無
	let scrollbar_v = (this.scrollbar === true || this.scrollbar === "v");
	let scrollbar_h = (this.scrollbar === true || this.scrollbar === "h");
	// スクロールバーが自動設定の場合
	if (this.scrollbar === "auto") {
		// 下記の計算を3回行う。
		// 1回目では、縦横スクロールバーが存在しない条件でのスクロールバーの有無を計算する
		// 2、3回目では、その前の計算によってスクロールバーが新たに存在することになった場合に、それを考慮に入れて計算をしなおす
		// 横が　無→有→有　縦が　無→無→有　のようになる可能性があるので、3回計算を行う必要がある。
		for (let i = 0 ; i < 3 ; i++) {
			// 縦スクロールバーが存在する場合で、横幅が自動設定でない場合、縦スクロールバーの分だけ、表示表行の横幅が狭まる
			// それを考慮に入れた場合の横スクロールバーの有無の計算
			if ((this.autosize === false || this.autosize === "height") && this.rect.w - (scrollbar_v ? this.scrollbar_data.v.width : 0) < this.rectall.w) {
				scrollbar_h = true;
			}
			// 上記の横スクロールバーバージョンの計算
			if ((this.autosize === false || this.autosize === "width") && this.rect.h - (scrollbar_h ? this.scrollbar_data.h.height : 0) < this.rectall.h) {
				scrollbar_v = true;
			}
		}
	}
	this.scrollbar_v = scrollbar_v;
	this.scrollbar_h = scrollbar_h;
	// 縦スクロールバーがありで、横幅が自動設定の場合は、スクロールバーの幅だけ横のサイズを伸ばす
	if ((this.autosize === true || this.autosize === "width") && scrollbar_v) {
		this.rect.w += this.scrollbar_data.v.width;
	}
	// 横スクロールバーがありで、縦幅が自動設定の場合は、スクロールバーの幅だけ縦サイズを伸ばす
	if ((this.autosize === true || this.autosize === "height") && scrollbar_h) {
		this.rect.h += this.scrollbar_data.h.height;
	}
	// スクロールバーに関するデータの設定
	if (scrollbar_v) {
		this.scrollbar_data.v.top = 0;
		this.scrollbar_data.v.rect.set(this.rect.x + this.rect.w - this.scrollbar_data.v.width, this.rect.y, this.scrollbar_data.v.width, this.rect.h - (scrollbar_h ? this.scrollbar_data.h.height: 0));
	}
	else {
		this.scrollbar_data.v.rect.set(0, 0, 0, 0);
	}
	if (scrollbar_h) {
		this.scrollbar_data.h.left = 0;
		this.scrollbar_data.h.rect.set(this.rect.x, this.rect.y + this.rect.h - this.scrollbar_data.h.height, this.rect.w - (scrollbar_v ? this.scrollbar_data.v.width: 0), this.scrollbar_data.h.height);
	}
	else {
		this.scrollbar_data.h.rect.set(0, 0, 0, 0);
	}
}

UI2.prototype.drawcommon = function() {
	if (!this.display) {
		return;
	}
	if (this.mousedownflag === "v") {
		let sdata = this.scrollbar_data.v;
		let diff = (this.mousedowny - sdata.tabrect.h / 2 - sdata.rect.y) - this.scrollbar_data.v.top;
		if (diff < -sdata.wheely) {
			diff = -sdata.wheely;
		}
		else if (diff > sdata.wheely) {
			diff = sdata.wheely;
		}
		this.scrollbar_data.v.top += diff;
	}
	else if (this.mousedownflag === "h") {
		let sdata = this.scrollbar_data.h;
		let diff = (this.mousedownx - sdata.tabrect.w / 2 - sdata.rect.x) - this.scrollbar_data.h.left;
		if (diff < -sdata.wheelx) {
			diff = -sdata.wheelx;
		}
		else if (diff > sdata.wheelx) {
			diff = sdata.wheelx;
		}
		this.scrollbar_data.h.left += diff;
	}
	if (this.mouseover) {
		let mouseover = false;
		if (this.type === "selector") {
			this.data.forEach(function(data, i, a) {
				if (data.mouseover) {
					data.mouseovercounter++;
					if (data.mouseovercounter >= data.tooltiptime && data.tooltip && this.enabletooltip) {
						mouseover = true;
						this.base.tooltipdata = data.tooltip.replace(/!S!/, data.shortcutkey);
						if (this.base.tooltip) {
							this.base.tooltip.mouseovercounter = 0;
						}
						this.base.tooltip = data;
						data.sx = this.sx;
						data.sy = this.sy;
						if (data.mouseovercounter == data.tooltiptime) {
							this.base.tooltipdatachanged = true;
						}
					}
				}
				else {
					data.mouseovercounter = 0;
				}
			}, this);
		}
		if (mouseover === false) {
			this.mouseovercounter++;
			if (this.mouseovercounter >= this.tooltiptime && this.tooltip && this.enabletooltip) {
				this.base.tooltipdata = this.tooltip.replace(/!S!/, this.shortcutkey);
				if (this.base.tooltip) {
					this.base.tooltip.mouseovercounter = 0;
				}
				this.base.tooltip = this;
				if (this.mouseovercounter == this.tooltiptime) {
					this.base.tooltipdatachanged = true;
				}
			}
		}
	}
	else {
		if (this.type === "selector") {
			this.data.forEach(function(data, i, a) {
				data.mouseovercounter = 0;
			}, this);
		}
		{
			this.mouseovercounter = 0;
		}
	}
	var ctx = this.ctx;
	var style = this.getstyle();
	if (style !== null) {
		if (style.fillStyle !== "transparent") {
			ctx.beginPath();
			ctx.rect(this.rect.x, this.rect.y, this.rect.w, this.rect.h);
			ctx.fillStyle = style.fillStyle;
			ctx.fill();
		}
	}
	ctx.save();
	ctx.beginPath();
	ctx.rect(this.rect.x, this.rect.y, this.rect.w - this.scrollbar_data.v.rect.w, this.rect.h - this.scrollbar_data.h.rect.h);
	ctx.clip();

	let scroll = this.scrollbar_h || this.scrollbar_v;
	if (scroll) {
		ctx.save();
		ctx.translate(this.scrollbar_h ? -this.scrollbar_data.h.left : 0, this.scrollbar_v ? -this.scrollbar_data.v.top : 0);
	}

	this.draw();

	if (scroll) {
		ctx.restore();
	}

	ctx.restore();
	// スクロールバーの描画
	let rect = this.rect;
	let rectall = this.rectall;
	if (!rectall.isempty()) {
		if (this.scrollbar_v) {
			let data = this.scrollbar_data.v;
			let srect = data.rect;
			let top = data.top;
			let bottom = top + srect.h;
			ctx.beginPath();
			ctx.fillStyle = "rgb(150, 150, 150)";
			ctx.rect(srect.x, srect.y, srect.w, srect.h);
			ctx.fill();
			if (bottom > rectall.h) {
				bottom = rectall.h;
				data.top = top = rectall.h - srect.h;
			}
			if (top < 0) {
				data.top = top = 0;
			}
			let tabrect = data.tabrect.set(srect.x, srect.y + Math.floor(top / rectall.h * srect.h), srect.w, Math.floor((bottom - top) / rectall.h * srect.h));
			ctx.beginPath();
			ctx.fillStyle = (this.mousedownflag === "tab_v" || this.mousedownflag === "v") ? "rgb(75, 75, 75)" : "rgb(100, 100, 100)";
			ctx.rect(tabrect.x, tabrect.y, tabrect.w, tabrect.h);
			ctx.fill();
		}
		if (this.scrollbar_h) {
			let data = this.scrollbar_data.h;
			let srect = data.rect;
			let left = data.left;
			let right = left + srect.w;
			ctx.beginPath();
			ctx.fillStyle = "rgb(150, 150, 150)";
			ctx.rect(srect.x, srect.y, srect.w, srect.h);
			ctx.fill();
			if (right > rectall.w) {
				right = rectall.w;
				data.left = left = rectall.w - srect.w;
			}
			if (left < 0) {
				data.left = left = 0;
			}
			let tabrect = data.tabrect.set(srect.x + Math.floor(left / rectall.w * srect.w), srect.y, Math.floor((right - left) / rectall.w * srect.w), srect.h);
			ctx.beginPath();
			ctx.fillStyle = (this.mousedownflag === "tab_h" || this.mousedownflag === "h") ? "rgb(75, 75, 75)" : "rgb(100, 100, 100)";
			ctx.rect(tabrect.x, tabrect.y, tabrect.w, tabrect.h);
			ctx.fill();
		}
	}

	if (style !== null) {
		if (style.lineWidth > 0) {
			ctx.beginPath();
			ctx.lineWidth = style.lineWidth;
			ctx.strokeStyle = style.strokeStyle;
			ctx.rect(this.rect.x + style.lineWidth / 2, this.rect.y + style.lineWidth / 2, this.rect.w - style.lineWidth, this.rect.h - style.lineWidth);
			ctx.strokeStyle = style.strokeStyle;
			ctx.stroke();
		}
	}

}
UI2.prototype.updatebindvar = function() {
}
UI2.prototype.preprocessorig = function() {
}
UI2.prototype.preprocess = function() {
}
UI2.prototype.checknewvalue = function() {
	return true;
}
UI2.prototype.onchange =  function() {
}
UI2.prototype.onchangeorig = function() {
	if (!this.isfirst) {
		this.onchange();
	}
}
UI2.prototype.onoperatedbyuser = function() {
}
UI2.prototype.draw =  function() {
	this.data.forEach(function(data, i, a) {
		data.on = this.on;
		data.mouseover = this.mouseover;
		let disabledbak = data.disabled;
		data.disabled = this.disabled || data.disabled;
		data.drawcommon(this.ctx);
		data.disabled = disabledbak;
	}, this);
}
UI2.prototype.paint = function(preprocessflag) {
	if (this.display === false) {
		return;
	}
	if (preprocessflag !== false) {
		this.base.preprocessall();
	}
	if (this === this.base) {
		this.tooltipdata = "";
		this.tooltip = null;
	}
	let ctx = this.ctx;
	ctx.fillStyle = "rgb(240, 240, 240)";
	ctx.beginPath();
	ctx.rect(this.rect.x, this.rect.y, this.rect.w, this.rect.h);
	ctx.fill();
	this.drawcommon();
	if (this === this.base && this.tooltipelement && this.enabletooltip) {
		if (this.tooltipdata) {
			if (this.tooltipdatachanged) {
				// ウィンドウ内の表示領域の長方形の位置、大きさを取得する
				let wleft = window.pageXOffset;
				let wtop = window.pageYOffset;
				let wwidth = document.documentElement.clientWidth;
				let wheight = document.documentElement.clientHeight;
				// ツールチップを左上の位置に配置した状態で表示状態で設定する
				this.tooltipelement.style.display = "block";
				this.tooltipelement.style.left = "0px";
				this.tooltipelement.style.right = "auto";
				this.tooltipelement.style.top = "0px";
				this.tooltipelement.style.bottom = "auto";
				this.tooltipelement.innerHTML = this.tooltipdata.replace(/(src=\')([^\']*\')/g, "$1" + this.path + "$2");
				// その表示幅、高さを取得する
				let width = this.tooltipelement.offsetWidth;
				let height = this.tooltipelement.offsetHeight;

				// ツールチップの表示位置の基準となるキャンバスの表示上の座標を取得する
				let crect = this.canvas.getBoundingClientRect();

				// ツールチップのデフォルトの表示位置を計算する
				// x座標はツールと同じ座標、y座標はツールの5ドット下
				let tleft = this.tooltip.rect.x + crect.left + wleft - this.tooltip.sx;
				let ttop = this.tooltip.rect.y + crect.top + wtop + this.tooltip.rect.h + 5 - this.tooltip.sy;
				let tright = wwidth - (tleft + width) - 1;
				let tbottom = wheight - (ttop + height) - 1;
				// 右にはみ出る場合は、右端から2ドットの位置に調整する
				if (tright < -wleft + 2) {
					tleft += tright + wleft - 2;
					tright = -wleft + 2;
				}
				// 左にはみ出るバイアは、左端から1ドットの位置に調整する
				if (tleft < wleft + 1) {
					tleft = wleft + 1;
				}

				// ツールチップの表示位置を設定する
				this.tooltipelement.style.left = tleft + "px";
				this.tooltipelement.style.right = tright + "px";
				// 下にはみでる場合で、上に表示できる場合は上に表示する
				if (tbottom < -wtop && ttop - wtop - this.tooltip.rect.h - 10 > height) {
					this.tooltipelement.style.top = (ttop - height - this.tooltip.rect.h - 10) + "px";
				}
				else {
					this.tooltipelement.style.top = ttop + "px";
				}
				this.tooltipdatachanged = false;
			}
		}
		else {
			this.tooltipelement.style.display = "none";
		}
	}
}
UI2.prototype.getdatafromid = function(id) {
	for (let i = 0 ; i < this.data.length ; i++) {
		if (this.data[i].id === id) {
			return this.data[i];
		}
	}
	return null;
}

function UI_Container(param) {
	var container = Object.create(UI_Container.prototype);
	Object.assign(container, UI2("container", param));
	container.space =					(param && param.space) || 0;
	container.lineheight =		(param && param.lineheight) || 0;
	container.linespacing =		(param && param.linespacing) || 0;
	container.update();
	return container;
}

UI_Container.prototype.update =	function() {
	if (this.type === "base") {
		this.rect.x = this.x;
		this.rect.y = this.y;
	}
	var x = this.rect.x + this.px;
	var y = this.rect.y + this.py;
	var maxx = x;
	var maxy = y;

	this.data.forEach(function(data, i, a) {
		if (data.position === "absolute") {
			data.rect.x = this.rect.x + data.x;
			data.rect.y = this.rect.y + data.y;
		}
		else if (data.position === "relative") {
			x += data.x;
			y += data.y;
			data.rect.x = x;
			data.rect.y = y;
		}
//		else if (data.position === "newline") {
//			x = this.rect.x + this.px + data.x;
//			y += data.y + this.lineheight;
//			data.rect.x = x;
//			data.rect.y = y;
//		}
		else if (data.position === "newline") {
			x = this.rect.x + this.px + data.x;
			y = maxy + data.y + this.linespacing;
			data.rect.x = x;
			data.rect.y = y;
		}
		else if (data.position === "newline_relative") {
			x = this.rect.x + this.px + data.x;
			y += data.y;
			data.rect.x = x;
			data.rect.y = y;
		}
		else {
			data.rect.x =		x;
			data.rect.y =		y;
		}
		if (maxx < data.rect.x + data.rect.w) {
			maxx = data.rect.x + data.rect.w;
		}
		if (maxy < data.rect.y + data.rect.h) {
			maxy = data.rect.y + data.rect.h;
		}
		data.ctx =		this.ctx;
		data.canvas =	this.canvas;
		data.arg =		this.arg;
		data.base	=		this.base;
		data.update();
		if (data.position !== "absolute") {
			x += data.rect.w + this.space;
		}
	}, this);
	this.setrect(maxx - this.rect.x, maxy - this.rect.y);
}
UI_Container.prototype.draw = function() {
	this.data.forEach(function(data, i, a) {
		let disabledbak = data.disabled;
		data.disabled = this.disabled || data.disabled;
		data.drawcommon(this.ctx);
		data.disabled = disabledbak;
	}, this);
}
UI_Container.prototype.preprocessorig = function() {
	var bindvar = this.bindvar;
	var bindvarvalue = bindvar && eval(bindvar);
	this.data.forEach(function(data, i, a) {
		if (bindvar) {
			data.display = bindvarvalue === data.value;
		}
	});
}
UI_Container.prototype.getdatafromid = function(id) {
	for (let i = 0 ; i < this.data.length ; i++) {
		if (this.data[i].id === id) {
			return this.data[i];
		}
		let retval = this.data[i].getdatafromid(id);
		if (retval) {
			return retval;
		}
	}
	return null;
}

Object.setPrototypeOf(UI_Container.prototype, UI2.prototype);

function UI_Base(param) {
	var container = Object.create(UI_Base.prototype);
	Object.assign(container, UI_Container(param));
	container.type = "base";
	container.base = container;
	if (container.canvas) {
		var canvas = container.canvas;
		canvas.ui = container;
		container.eventorig = {};
		UI2.eventlist.forEach(function(data, i, a) {
			canvas.addEventListener(data, container[data + "base"]);
		}, this);
	}
	container.update();
	container.path = (param && param.path) || "";
	container.tooltipelement = (param && param.tooltipelement) || null;
	return container;
}

UI2.eventlist.forEach(function(data, i, a) {
	UI_Base.prototype[data + "base"] = function(evt) {
		// evt.deltaX は freeze されていて変更できないのでコピーする
		evt.mydeltaX = evt.deltaX;
		evt.mydeltaY = evt.deltaY;
		// スクロールによってどれだけずれたか
		evt.sx = 0;
		evt.sy = 0;
		var brect = this.getBoundingClientRect();
		// マウスの座標
		evt.mx = evt.mx2 = Math.floor(evt.clientX - brect.left);
		evt.my = evt.my2 = Math.floor(evt.clientY - brect.top);
		this.ui.mouseover = this.ui.rect.isinside(evt.mx, evt.my);

		this.ui[data + "preprocess"](evt);
		if (!this.ui.disabled) {
			this.ui[data](evt);
			this.ui["on" + data].apply(this.ui, [ evt ]);
		}
		this.ui.paint(false);
	};

	UI2.checkshortcut = function(evt, shortcutkey) {
		let keycode = shortcutkey.substr(0, 1).toUpperCase().charCodeAt(0);
		let altkey = false;
		let ctrlkey = false;
		let functionkey = false;
		if (shortcutkey.substr(0, 5).toLowerCase() === "ctrl+") {
			ctrlkey = true;
			keycode = shortcutkey.substr(5, 1).toUpperCase().charCodeAt(0);
		}
		else if (shortcutkey.substr(0, 4).toLowerCase() === "alt+") {
			altkey = true;
			keycode = shortcutkey.substr(4, 1).toUpperCase().charCodeAt(0);
		}
		if (ctrlkey == evt.ctrlKey && altkey == evt.altKey && evt.keyCode === keycode) {
			return true;
		}
		else {
			return false;
		}
	}

	UI2.prototype.preprocessall = function() {
		this.preprocess();
		if (!this.display) {
			return;
		}
		this.isfirst = false;
		this.preprocessorig();
		this.data.forEach(function(data2, i, a) {
			if (this.type === "container" || this.type === "base") {
				data2.preprocessall();
			}
		}, this);
	}

	UI2.prototype[data + "preprocess"] = function(evt) {
		this.preprocess();
		if (!this.display) {
			return;
		}

		let isinside = this.rect.isinside(evt.mx, evt.my);
		this.sx = evt.sx;
		this.sy = evt.sy;
		this.preprocessorig();
		let mxbak = evt.mx;
		let mybak = evt.my;
		if (data === "wheel") {
			let sdatav = this.scrollbar_data.v;
			let sdatah = this.scrollbar_data.h;
			let dy = sdatav.wheely;
			if (dy <= 0) {
				dy = 1;
			}
			let dx = sdatah.wheelx;
			if (dx <= 0) {
				dx = 1;
			}
			if (sdatav.rect.isinside(evt.mx, evt.my)) {
				if (evt.mydeltaY > 0 || evt.mydeltaX > 0) {
					sdatav.top += dy;
					evt.mydeltaX = 0;
					evt.mydeltaY = 0;
				}
				else if (evt.mydeltaY < 0 || evt.mydeltaX < 0) {
					sdatav.top -= dy;
					evt.mydeltaX = 0;
					evt.mydeltaY = 0;
				}
				evt.mx = -100000000;
				evt.my = -100000000;
			}
			else if (sdatah.rect.isinside(evt.mx, evt.my)) {
				if (evt.mydeltaY > 0 || evt.mydeltaX > 0) {
					sdatah.left += dx;
					evt.mydeltaX = 0;
					evt.mydeltaY = 0;
				}
				else if (evt.mydeltaY < 0 || evt.mydeltaX < 0) {
					sdatah.left -= dx;
					evt.mydeltaX = 0;
					evt.mydeltaY = 0;
				}
				evt.mx = -100000000;
				evt.my = -100000000;
			}
			else if (isinside && this.enablescrollwheel) {
				if (this.scrollbar_v) {
					if (evt.mydeltaY > 0) {
						sdatav.top += dy;
					}
					else if (evt.mydeltaY < 0) {
						sdatav.top -= dy;
					}
				}
				if (this.scrollbar_h) {
					if (evt.mydeltaX > 0) {
						sdatah.left += dx;
					}
					else if (evt.mydeltaX < 0) {
						sdatah.left -= dx;
					}
				}
			}
		}
		else if (this.mousedownflag) {
			if (data === "mousemove") {
				if (this.mousedownflag === "tab_v") {
					this.scrollbar_data.v.top = this.mousedowntop + evt.my2 - this.mousedowny;
				}
				else if (this.mousedownflag === "tab_h") {
					this.scrollbar_data.h.left = this.mousedownleft + evt.mx2 - this.mousedownx;
				}
				else if (this.mousedownflag === "v") {
					this.mousedowny = evt.my;
				}
				else if (this.mousedownflag === "h") {
					this.mousedownx = evt.mx;
				}
			}
			else if (data === "mouseup" || data === "mouseout") {
				this.mousedownflag = false;
			}
			evt.mx = -100000000;
			evt.my = -100000000;
		}
		else if ((data === "mousedown" || data === "mouseup" || data === "mousemove")) {
			if (this.scrollbar_data.v.rect.isinside(evt.mx, evt.my)) {
				if (data === "mousedown") {
					if (this.scrollbar_data.v.tabrect.isinside(evt.mx, evt.my)) {
						this.mousedownflag = "tab_v";
						this.mousedowntop = this.scrollbar_data.v.top;
						this.mousedowny = evt.my;
					}
					else {
						this.mousedownflag = "v";
						this.mousedowny = evt.my;
					}
				}
				else if (data === "mouseup") {
					this.mousedownflag = false;
				}
				evt.mx = -100000000;
				evt.my = -100000000;
			}
			else if (this.scrollbar_data.h.rect.isinside(evt.mx, evt.my)) {
				if (data === "mousedown") {
					if (this.scrollbar_data.h.tabrect.isinside(evt.mx, evt.my)) {
						this.mousedownflag = "tab_h";
						this.mousedownleft = this.scrollbar_data.h.left;
						this.mousedownx = evt.mx;
					}
					else {
						this.mousedownflag = "h";
						this.mousedownx = evt.mx;
					}
				}
				else if (data === "mouseup") {
					this.mousedownflag = false;
				}
				evt.mx = -100000000;
				evt.my = -100000000;
			}
		}
		if (this.scrollbar_h) {
			evt.mx += this.scrollbar_data.h.left;
			evt.mx2 += this.scrollbar_data.h.left;
			evt.sx += this.scrollbar_data.h.left;
		}
		if (this.scrollbar_v) {
			evt.my += this.scrollbar_data.v.top;
			evt.my2 += this.scrollbar_data.v.top;
			evt.sy += this.scrollbar_data.v.top;
		}

		if (!isinside) {
			evt.mx = -100000000;
			evt.my = -100000000;
		}
		if (data === "keydown" && this.enableshortcut) {
			if (this.type === "selector") {
				this.data.forEach(function(data, i, a) {
					if (data.shortcutkey && UI2.checkshortcut(evt, data.shortcutkey)) {
						this.selectnum = i;
						this.selectvalue = this.data[i].value;
						this.updatebindvar();
						this.onchangeorig();
						this.onoperatedbyuser();
						return;
					}
				}, this);
			}
			else if (this.type === "clickselector") {
				if (this.shortcutkey && UI2.checkshortcut(evt, this.shortcutkey)) {
					evt.buttons |= 1;
					this.mousedown(evt);
					this.onmousedown(evt);
					this.onoperatedbyuser();
				}
			}
			else {
				if (this.shortcutkey && UI2.checkshortcut(evt, this.shortcutkey)) {
					evt.buttons |= 1;
					this.onmousedown(evt);
					this.onoperatedbyuser();
				}
			}
		}
		this.isfirst = false;
		this.data.forEach(function(data2, i, a) {
			if (evt) {
				data2.mouseover = data2.rect.isinside(evt.mx, evt.my);
			}
			if (this.type === "container" || this.type === "base") {
				data2[data + "preprocess"](evt);
				if (data2.display && evt && !data2.disabled) {
					if (data === "wheel" || data2.mouseover) {
						data2[data](evt);
						data2["on" + data](evt);
						data2["post" + data](evt);
					}
				}
			}
		}, this);
		evt.mx = mxbak;
		evt.my = mybak;
	};
	UI2.prototype[data] = function(evt) {
	};
	UI2.prototype["on" + data] = function(evt) {
	};
	UI2.prototype["post" + data] = function(evt) {
	};
});

Object.setPrototypeOf(UI_Base.prototype, UI_Container.prototype);

function UI_Button(param) {
	var button = Object.create(UI_Button.prototype);
	Object.assign(button, UI2("button", param));
	if (!param || !param.moveroffstyle) {
		button.moveroffstyle = button.moverstyle;
	}
	button.update();
	return button;
}

UI_Button.prototype.update = function() {
	this.data.forEach(function(data, i, a) {
		data.arg 		=	this.arg;
		data.ctx		=	this.ctx;
		data.base		= this.base;
		data.rect.x =	this.rect.x + this.px;
		data.rect.y =	this.rect.y + this.py;
		data.update();
	}, this);
	this.setrect(this.px * 2 + this.data[0].rect.w, this.py * 2 + this.data[0].rect.h);
}
UI_Button.prototype.postmousedown = function(evt) {
	if (evt.buttons & 1) {
		this.onoperatedbyuser();
	}
}
Object.setPrototypeOf(UI_Button.prototype, UI2.prototype);

function UI_SelectorCommon(type, param) {
	var selector = Object.create(UI_SelectorCommon.prototype);
	Object.assign(selector, UI2(type, param));
	selector.onchangesavevalue =	(param && param.onchangesavevalue !== undefined) ? param.onchangesavevalue : false;
	selector.selectnum =					(param && param.selectnum) || 0;
	selector.rotatebywheel =	(param && param.rotatebywheel !== undefined) ? param.rotatebywheel : true;
	selector.selectvalue =				"none";
	return selector;
}
UI_SelectorCommon.prototype.preprocessorig = function() {
	if (this.arg && this.bindvar) {
		var selectvalue = eval(this.bindvar);
		let selectnumorig = this.selectnum;
		var selectnum = -1;
		this.data.forEach(function(data, i, a) {
			if (data.value === selectvalue) {
				selectnum = i;
				return;
			}
		});
		if (selectnum >= 0) {
			if (selectnum != selectnumorig) {
				this.selectvalue = selectvalue;
				this.selectnum = selectnum;
				this.onchangeorig();
			}
		}
		else {
			if (this.allowunselect) {
				if (selectnum != selectnumorig) {
					this.selectvalue = selectvalue;
					this.selectnum = selectnum;
					this.onchangeorig();
				}
			}
			else {
				this.updatebindvar();
			}
		}
	}
}
UI_SelectorCommon.prototype.updatebindvar = function() {
	if (this.arg && this.bindvar) {
		eval(this.bindvar + " = this.selectvalue;");
	}
}

UI_SelectorCommon.prototype.checkselect = function(type) {
	if (type === "isfirst") {
		for (let i = 0 ; i < this.data.length ; i++) {
			if (!this.data[i].disabled) {
				if (this.selectnum === i) {
					return true;
				}
				else {
					return false;
				}
			}
		}
		return false;
	}
	else if (type === "islast") {
		for (let i = this.data.length - 1 ; i >= 0 ; i--) {
			if (!this.data[i].disabled) {
				if (this.selectnum === i) {
					return true;
				}
				else {
					return false;
				}
			}
		}
		return false;
	}
}

UI_SelectorCommon.prototype.select = function(type) {
	this.overflow = false;
	this.underflow = false;
	this.changed = false;
	let prevselectnum = this.selectnum;
	if (type === "first") {
		for (let i = 0 ; i < this.data.length ; i++) {
			if (!this.data[i].disabled) {
				this.selectnum = i;
				break;
			}
		}
	}
	else if (type === "last") {
		for (let i = this.data.length - 1 ; i >= 0 ; i--) {
			if (!this.data[i].disabled) {
				this.selectnum = i;
				break;
			}
		}
	}
	else if (type === "next") {
		let i = 0;
		while (i < this.data.length) {
			if (this.selectnum == this.data.length - 1) {
				this.overflow = true;
			}
			if (!this.rotatevaluebywheel && this.selectnum == this.data.length - 1) {
				this.selectnum = prevselectnum;
				break;
			}
			this.selectnum = (this.selectnum + 1) % this.data.length;
			if (this.data[this.selectnum].display && !this.data[this.selectnum].disabled) {
				break;
			}
			if (prevselectnum === this.selectnum) {
				break;
			}
			i++;
		}
	}
	else if (type === "prev") {
		let i = 0;
		while (i < this.data.length) {
			if (this.selectnum == 0) {
				this.underflow = true;
			}
			if (!this.rotatevaluebywheel && this.selectnum == 0) {
				this.selectnum = prevselectnum;
				break;
			}
			this.selectnum = (this.selectnum + this.data.length - 1) % this.data.length;
			if (this.data[this.selectnum].display && !this.data[this.selectnum].disabled) {
				break;
			}
			i++;
		}
	}
	if (prevselectnum !== this.selectnum) {
		this.changed = true;
		this.selectvalue = this.data[this.selectnum].value;
		this.updatebindvar();
		this.onchangeorig();
		this.onoperatedbyuser();
	}
}

UI_SelectorCommon.prototype.wheel = function(evt) {
	this.overflow = false;
	this.underflow = false;
	if (this.data.length > 0) {
		if (this.enablewheel === true || (this.enablewheel === "mouseover" && this.rect.isinside(evt.mx, evt.my))) {
			if (evt.mydeltaY > 0) {
				this.select("next");
			}
			else if (evt.mydeltaY < 0) {
				this.select("prev");
			}
			else {
				return;
			}
		}
	}
}
// todo disabled 対応
UI_SelectorCommon.prototype.selectbyvalue = function(value) {
	let selectnumorig = this.selectnum;
	let selectnum = -1;
	this.data.forEach(function(data, i, a) {
		if (data.value === value) {
			selectnum = i;
			return;
		}
	}, this);
	if (selectnum >= 0) {
		if (selectnum != selectnumorig) {
			this.selectvalue = value;
			this.selectnum = selectnum;
			this.updatebindvar();
			this.onchangeorig();
		}
	}
	else {
		if (this.allowunselect) {
			if (selectnum != selectnumorig) {
				this.selectvalue = value;
				this.selectnum = selectnum;
				this.updatebindvar();
				this.onchangeorig();
			}
		}
	}
}
UI_SelectorCommon.prototype.onchangeorig = function(value) {
	if (!this.isfirst) {
		this.onchange();
	}
	if (this.autoscroll && this.selectnum >= 0) {
		let data = this.data[this.selectnum];
		if (this.scrollbar_v) {
			let sdata = this.scrollbar_data.v;
			let top = sdata.rect.y + sdata.top;
			let bottom = top + sdata.rect.h;
			if (data.rect.y + data.rect.h > bottom) {
				bottom = data.rect.y + data.rect.h;
				top = bottom - sdata.rect.h;
			}
			if (data.rect.y < top) {
				top = data.rect.y;
				bottom + top + sdata.rect.h;
			}
			sdata.top = top - sdata.rect.y;
		}
		if (this.scrollbar_h) {
			let sdata = this.scrollbar_data.h;
			let left = sdata.rect.x + sdata.left;
			let right = left + sdata.rect.w;
			if (data.rect.x + data.rect.w > right) {
				right = data.rect.x + data.rect.w;
				left = right - sdata.rect.w;
			}
			if (data.rect.x < left) {
				left = data.rect.x;
				right + top + sdata.rect.w;
			}
			sdata.left = left - sdata.rect.x;
		}
	}
}
Object.setPrototypeOf(UI_SelectorCommon.prototype, UI2.prototype);

function UI_Selector(param) {
	var selector = Object.create(UI_Selector.prototype);
	Object.assign(selector, UI_SelectorCommon("selector", param));
	selector.space =				(param && param.space) || 0;
	selector.lineheight =		(param && param.lineheight) || 10;
	selector.linespacing =		(param && param.linespacing) || 0;
	selector.allowunselect =	(param && param.allowunselect) || false;
	selector.autoscroll = (param && param.autoscroll !== undefined) ? param.autoscroll :true;
	selector.update();
	return selector;
}

UI_Selector.prototype.update= function() {
	var x =		this.rect.x + this.px;
	var y =		this.rect.y + this.py;
	var maxx	= x;
	var maxy	= y;
	var selectnum = +this.selectnum;
	if (selectnum < 0 || selectnum >= this.data.length) {
		selectnum = this.selectnum = 0;
	}
	if (this.data.length > 0) {
		this.selectvalue = this.data[selectnum].value;
	}
	else {
		this.selectvalue = "";
	}

	this.data.forEach(function(data, i, a) {
		if (data.position === "absolute") {
			data.rect.x = this.rect.x + data.x;
			data.rect.y = this.rect.y + data.y;
		}
		else if (data.position === "relative") {
			x += data.x;
			y += data.y;
			data.rect.x = x;
			data.rect.y = y;
		}
//		else if (data.position === "newline") {
//			x = this.rect.x + this.px + data.x;
//			y += data.y + this.lineheight;
//			data.rect.x = x;
//			data.rect.y = y;
//		}
		else if (data.position === "newline") {
			x = this.rect.x + this.px + data.x;
			y = maxy + data.y + this.linespacing;
			data.rect.x = x;
			data.rect.y = y;
		}
		else if (data.position === "newline_relative") {
			x = this.rect.x + this.px + data.x;
			y += data.y;
			data.rect.x = x;
			data.rect.y = y;
		}
		else {
			data.rect.x =		x;
			data.rect.y =		y;
		}
		if (maxx < data.rect.x + data.rect.w) {
			maxx = data.rect.x + data.rect.w;
		}
		if (maxy < data.rect.y + data.rect.h) {
			maxy = data.rect.y + data.rect.h;
		}

		data.ctx =	this.ctx;
		data.arg =	this.arg;
		data.base = this.base;
		data.update();
		data.on = (i == selectnum);
		if (data.position !== "absolute") {
			x += data.rect.w + this.space;
		}
	}, this);
	this.setrect(maxx - this.rect.x, maxy - this.rect.y);
}
UI_Selector.prototype.draw =function() {
	let selectnum = this.selectnum;
	this.data.forEach(function(data, i, a) {
		data.on = (selectnum == i);
		if (data.display) {
			let disabledbak = data.disabled;
			data.disabled = this.disabled || data.disabled;
			data.drawcommon(this.ctx);
			data.disabled = disabledbak;
		}
	}, this);
}
UI_Selector.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	var newselectnum = -1;
	this.data.forEach(function(data, i, a) {
		if (data.display && data.mouseover && !data.disabled) {
			newselectnum = i;
			return;
		}
	}, this);
	if (newselectnum != -1 && (this.onchangesavevalue || newselectnum != this.selectnum)) {
		this.selectnum = newselectnum;
		this.selectvalue = this.data[this.selectnum].value;
		this.updatebindvar();
		this.onchangeorig();
		this.onoperatedbyuser();
	}
}

Object.setPrototypeOf(UI_Selector.prototype, UI_SelectorCommon.prototype);

function UI_ClickSelector(param) {
	var selector = Object.create(UI_ClickSelector.prototype);
	Object.assign(selector, UI_SelectorCommon("clickselector", param));
	selector.selectnum =		(param && param.selectnum) || 0;
	selector.selectvalue = null;
	if (!param || !param.moveroffstyle) {
		selector.moveroffstyle = selector.moverstyle;
	}
	selector.update();
	return selector;
}

UI_ClickSelector.prototype.update = function() {
	var x =		this.rect.x + this.px;
	var y =		this.rect.y + this.py;
	var selectnum = +this.selectnum;
	if (selectnum < 0 || selectnum >= this.data.length) {
		selectnum = this.selectnum = 0;
	}
	if (this.data.length > 0) {
		this.selectvalue = this.data[selectnum].value;
	}
	else {
		this.selectvalue = null;
	}
	var maxw = 0;
	var maxh = 0;
	this.data.forEach(function(data, i, a) {
		data.rect.x =		x;
		data.rect.y =		y;
		data.arg =			this.arg;
		data.base =			this.base;
		if (maxw < data.rect.w) {
			maxw = data.rect.w;
		}
		if (maxh < data.rect.h) {
			maxh = data.rect.h;
		}
		data.update();
	});
	this.setrect(maxw + this.px * 2, maxh + this.py * 2);
}
UI_ClickSelector.prototype.draw = function(ui) {
	if (this.data.length > 0) {
		this.data[this.selectnum].drawcommon(this.ctx);
	};
}
UI_ClickSelector.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	// todo disabled 対応
	if (this.data.length > 0) {
		if (evt.buttons & 1) {
			this.selectnum = (this.selectnum + 1) % this.data.length;
		}
		else if (evt.buttons & 2) {
			this.selectnum = (this.selectnum + this.data.length - 1) % this.data.length;
		}
		this.selectvalue = this.data[this.selectnum].value;
		this.updatebindvar();
		this.onchangeorig();
		this.onoperatedbyuser();
	}
}

Object.setPrototypeOf(UI_ClickSelector.prototype, UI_SelectorCommon.prototype);

function UI_TextBox(param) {
	var textbox = Object.create(UI_TextBox.prototype);
	Object.assign(textbox, UI2("textbox", param));
	if (!param || !param.moveroffstyle) {
		textbox.moveroffstyle = textbox.moverstyle;
	}
	textbox.value		= (param && param.value) || "";
	textbox.maxlen	= (param && param.maxlen) || 16;
	textbox.msg			= (param && param.msg) || "";
	textbox.update();
	return textbox;
}

UI_TextBox.prototype.update = function() {
	if (this.data.length > 0) {
		this.data[0].arg =		this.arg;
		this.data[0].base =		this.base;
		this.data[0].rect.x =	this.rect.x + this.px;
		this.data[0].rect.y =	this.rect.y + this.py;
		this.data[0].update();
		this.setrect(this.px * 2 + this.data[0].rect.w, this.py * 2 + this.data[0].rect.h);
	}
}
UI_TextBox.prototype.preprocessorig = function() {
	if (this.arg && this.bindvar) {
		this.value = eval(this.bindvar);
	}
	if (this.data.length > 0) {
			this.data[0].text = this.value;
	}
}
UI_TextBox.prototype.updatebindvar = function() {
	if (this.arg && this.bindvar) {
		eval(this.bindvar + " = this.value;");
	}
}
UI_TextBox.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	let text = window.prompt(this.msg, this.value);
	if (text !== null) {
		this.changevalue(text);
		this.onoperatedbyuser();
	}
}

UI_TextBox.prototype.changevalue = function(value, redrawflag) {
	this.newvalue = value;
	this.prevvalue = this.value;
	if (this.newvalue === this.value) {
		return;
	}
	if (this.newvalue.length > this.maxlen) {
		this.newvalue = this.newvalue.substr(0, this.maxlen);
	}
	this.value = this.newvalue;
	this.updatebindvar();
	if (this.data.length > 0) {
		this.data[0].text = this.value;
	}
	this.onchangeorig();
	if (redrawflag) {
		this.draw();
	}
}

Object.setPrototypeOf(UI_TextBox.prototype, UI2.prototype);

function UI_NumberBox(param) {
	var numbox = Object.create(UI_NumberBox.prototype);
	Object.assign(numbox, UI2("numberbox", param));
	if (!param || !param.moveroffstyle) {
		numbox.moveroffstyle = numbox.moverstyle;
	}
	numbox.value			= (param && param.value) || 0;
	numbox.msg				= (param && param.msg) || "";
	numbox.numtype		= (param && param.numtype) || "real";
	numbox.max				= (param && param.max !== undefined) ? param.max : Infinity;
	numbox.min				= (param && param.min !== undefined) ? param.min : -Infinity;
	numbox.wheelnum		= (param && param.wheelnum) || 1;
	numbox.update();
	return numbox;
}

UI_NumberBox.prototype.update = function() {
	if (this.data.length > 0) {
		this.data[0].arg =		this.arg;
		this.data[0].base =		this.base;
		this.data[0].rect.x =	this.rect.x + this.px;
		this.data[0].rect.y =	this.rect.y + this.py;
		this.data[0].text =		this.value;
		this.data[0].update();
		this.setrect(this.px * 2 + this.data[0].rect.w, this.py * 2 + this.data[0].rect.h);
	}
}

UI_NumberBox.prototype.changevalue = function(value, redrawflag) {
	this.newvalue = value;
	this.prevvalue = this.value;
	if (this.numtype === "integer") {
		this.newvalue	|= 0;
		this.min			|= 0;
		this.max			|= 0;
	}
	else {
		this.newvalue	= +value;
		this.min			= +this.min;
		this.max			= +this.max;
	}
	if (this.min > this.max) {
		this.min = this.max;
	}
	if (this.newvalue < this.min) {
		this.newvalue = this.min;
	}
	else if (this.newvalue > this.max) {
		this.newvalue = this.max;
	}
	if (this.newvalue === this.value) {
		return;
	}
	if (this.checknewvalue()) {
		this.value = this.newvalue;
		this.updatebindvar();
		if (this.data.length > 0) {
				this.data[0].text = this.value;
		}
		this.onchangeorig();
	}
	if (redrawflag) {
		this.draw();
	}
}

UI_NumberBox.prototype.preprocessorig = function() {
	let newvalue = this.value;
	if (this.arg && this.bindvar) {
		newvalue = eval(this.bindvar);
	}
	if (newvalue !== this.value) {
		this.changevalue(newvalue);
	}
}
UI_NumberBox.prototype.updatebindvar = function() {
	if (this.arg && this.bindvar) {
		eval(this.bindvar + " = this.value;");
	}
}
UI_NumberBox.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	let num = window.prompt(this.msg, this.value);
	if (num !== null) {
		if (this.numtype === "integer") {
			num = num | 0;
		}
		else {
			num = +num;
		}
		this.changevalue(num);
		this.onoperatedbyuser();
	}
}
UI_NumberBox.prototype.wheel = function(evt) {
	this.overflow = false;
	this.underflow = false;
	if (this.enablewheel === true || (this.enablewheel === "mouseover" && this.rect.isinside(evt.mx, evt.my))) {
		if (evt.mydeltaY > 0) {
			if (this.value == this.max) {
				this.overflow = true;
			}
			this.changevalue(this.value + this.wheelnum);
			this.onoperatedbyuser();
		}
		else if (evt.mydeltaY < 0) {
			if (this.value == this.min) {
				this.underflow = true;
			}
			this.changevalue(this.value - this.wheelnum);
			this.onoperatedbyuser();
		}
		else {
			return;
		}
	}
}

Object.setPrototypeOf(UI_NumberBox.prototype, UI2.prototype);



